%

%
% Copyright (c) 2016 LAAS/CNRS
% All rights reserved.
%
% Redistribution  and  use  in  source  and binary  forms,  with  or  without
% modification, are permitted provided that the following conditions are met:
%
%   1. Redistributions of  source  code must retain the  above copyright
%      notice and this list of conditions.
%   2. Redistributions in binary form must reproduce the above copyright
%      notice and  this list of  conditions in the  documentation and/or
%      other materials provided with the distribution.
%
% THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
% WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
% MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
% ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
% WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
% IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
%
%                                           Anthony Mallet on Wed Mar  2 2016
%
function genomix_mask(block, what)
  switch what
    case 'init'
      init(block);

    case 'fetch'
      fetch(block);
    case 'port'
      portsel(block);
    case 'signame'
      signamesel(block);

    case 'add'
      add(block);
    case 'remove'
      remove(block);

    case 'timetype'
      timetype(block);

    case 'uipath'
      path = get_param(block, 'path');
      [file, path, ~] = uigetfile('*.so', 'Select a folder', path);
      if ischar(path)
        set_param(block, 'path', path);
        [~, comp, ~] = fileparts(file);
        set_param(block, 'component', comp);

        fetch(block);
      end
  end
end

function init(block)
  set_param(block, 'userdatapersistent', 'on');
  update(block);

  kind = get_param(block, 'kind');
  udata = get_param(block, 'userdata');

  switch kind
    case 'port'
      if isfield(udata, 'meta') && isfield(udata.meta, 'ports')
        port = get_param(block, 'port');
        sub = get_param(block, 'subport');
        if ~isempty(port) && udata.meta.ports.(port).simple
          if ~isempty(sub)
            error('port %s cannot have a sub-component name', port);
          end
        else
          if isempty(sub)
            error('port %s must have a sub-component name', port);
          end
        end
    end
  end

  if isempty(udata.siglist)
    md = 'disp(''(not configured)'')';
  else
    if strcmp(kind, 'port'); dir = 'output'; else dir = 'input'; end
    md = '';
    i = 1;
    for s = udata.siglist
      d = s.name;
      if strcmp(s.type, 'size'); d = [d ' (size)']; end
      md = [md sprintf('port_label(''%s'',%d,''%s'')\n', dir, i, d)];
      i = i+1;
    end

    ttype = get_param(block, 'timetype');
    if ~strcmp(ttype, 'periodic')
      if strcmp(kind, 'port'); i = 1; end

      ph = get_param(block, 'porthandles');
      if i <= length(ph.Inport)
        md = [md sprintf('port_label(''input'',%d,''EN'')\n', i)];
      end
    end
  end

  instance = get_param(block, 'instance');
  if isempty(instance)
    instance = get_param(block, 'component');
  end
  port = get_param(block, 'port');
  fs = sprintf('%s/%s', instance, port);
  if strcmp(fs, '/'); fs = ''; end

  set_param(block, 'attributesformatstring', fs, 'maskdisplay', md);
end

function update(block)
  udata = get_param(block, 'userdata');
  if ~isfield(udata, 'meta')
    udata.meta = struct();
    udata.siglist = [];
    udata.signame = {};
    udata.sigsize = containers.Map();
    set_param(block, 'userdata', udata);
  end

  sl = '';
  for s = udata.siglist
    sl = [sl sprintf('{"name":"%s","type":"%s","size":%d},', ...
                     s.name, s.type, s.size)];
  end
  set_param(block, 'siglist', ['[' sl(1:end-1) ']']);

  if isempty(udata.siglist)
     sd = '(empty)';
  else
    sd = '';
    for s = udata.siglist
      sd = [sd sprintf('%c %-25s\t%s (dimension %d)\n', ...
                       8226, s.name, s.type, s.size)];
    end
  end
  mask = Simulink.Mask.get(block);
  d = mask.getDialogControl('descr');
  d.Prompt = sd;
end


% --- fetch ----------------------------------------------------------------
%
% Load component, get ports and services
%
function fetch(block)
  component = get_param(block, 'component');
  if isempty(component)
    errordlg('empty component name', 'genomix', 'modal');
    return;
  end
  kind = get_param(block, 'kind');
  host = get_param(block, 'host');
  path = get_param(block, 'path');
  port = get_param(block, 'port');
  instance = get_param(block, 'instance');
  udata = get_param(block, 'userdata');

  mask = Simulink.Mask.get(block);

  wait = waitbar(0, 'Please wait...');
  try
    waitbar(0.3, wait);
    h = genomix.client(host);

    waitbar(0.6, wait);
    if ~isempty(path)
      component = [path '/' component];
    end
    if isempty(instance)
      c = h.load(component);
    else
      c = h.load(component, '--name', instance);
    end
    udata.meta = c.meta;

    waitbar(0.9, wait);
    switch kind
      case 'port'
        portname = fieldnames(udata.meta.ports);
      case 'rqst'
        portname = fieldnames(udata.meta.services);
    end
    if ~ismember(port, portname)
      port = portname{1};
      set_param(block, 'port', port);
      switch kind
        case 'port'
          set_param(block, 'subport', '');
      end
    end

    switch kind
      case 'port'
        i = find(strcmp({mask.Parameters.Name}, 'subport'));
        if udata.meta.ports.(port).simple
          mask.Parameters(i).Enabled = 'off';
        else
          mask.Parameters(i).Enabled = 'on';
        end
    end

    udata.siglist = [];
    switch kind
      case 'port'
        meta = udata.meta.ports.(port).data;
      case 'rqst'
        meta = udata.meta.services.(port).input;
    end
    [udata.signame, udata.sigsize] = signames(meta);

  catch e
    udata.meta = struct();
    udata.siglist = [];
    udata.sigsize = containers.Map();
    udata.signame = {};
    portname = {};
    port = '';
    set_param(block, 'port', '');

    errordlg(e.message, 'genomix', 'modal');
  end

  i = find(strcmp({mask.Parameters.Name}, 'portsel'));
  if ~isequal(mask.Parameters(i).TypeOptions, portname)
    mask.Parameters(i).TypeOptions = portname;
    if ~isempty(port)
      mask.Parameters(i).Value = port;
    end
  end

  i = find(strcmp({mask.Parameters.Name}, 'signame'));
  if ~isequal(mask.Parameters(i).TypeOptions, udata.signame)
    mask.Parameters(i).TypeOptions = udata.signame;
  end

  set_param(block, 'userdata', udata);
  update(block)
  close(wait);
end


% --- portsel --------------------------------------------------------------
%
% Update port selection
%
function portsel(block)
  port = get_param(block, 'port');
  portsel = get_param(block, 'portsel');
  if strcmp(portsel, port); return; end

  kind = get_param(block, 'kind');
  udata = get_param(block, 'userdata');
  mask = Simulink.Mask.get(block);

  if strcmp(portsel, '<empty>')
    if isempty(port); return; end;
    if ~isfield(udata, 'meta') || isempty(udata.meta); return; end

    switch kind
      case 'port'
        portname = fieldnames(udata.meta.ports);

        i = find(strcmp({mask.Parameters.Name}, 'subport'));
        if udata.meta.ports.(port).simple
          mask.Parameters(i).Enabled = 'off';
          set_param(block, 'subport', '');
        else
          mask.Parameters(i).Enabled = 'on';
        end

      case 'rqst'
        portname = fieldnames(udata.meta.services);
    end

    i = find(strcmp({mask.Parameters.Name}, 'portsel'));
    mask.Parameters(i).TypeOptions = portname;
    mask.Parameters(i).Value = port;
    return
  end

  % update port
  set_param(block, 'port', portsel);

  udata.siglist = [];
  switch kind
    case 'port'
      meta = udata.meta.ports.(portsel).data;
      i = find(strcmp({mask.Parameters.Name}, 'subport'));
      if udata.meta.ports.(portsel).simple
        mask.Parameters(i).Enabled = 'off';
        set_param(block, 'subport', '');
      else
        mask.Parameters(i).Enabled = 'on';
      end

    case 'rqst'
      meta = udata.meta.services.(portsel).input;
  end
  [udata.signame, udata.sigsize] = signames(meta);

  i = find(strcmp({mask.Parameters.Name}, 'signame'));
  if ~isequal(mask.Parameters(i).TypeOptions, udata.signame)
    mask.Parameters(i).TypeOptions = udata.signame;
  end

  set_param(block, 'userdata', udata);
  update(block)
end


% --- signamesel -----------------------------------------------------------
%
% Update signame selection
%
function signamesel(block)
  signamesel = get_param(block, 'signame');

  if strcmp(signamesel, '<empty>')
    udata = get_param(block, 'userdata');
    if ~isfield(udata, 'signame') || isempty(udata.signame); return; end

    mask = Simulink.Mask.get(block);
    i = find(strcmp({mask.Parameters.Name}, 'signame'));
    if ~isequal(mask.Parameters(i).TypeOptions, udata.signame)
      mask.Parameters(i).TypeOptions = udata.signame;
    end
  end
end


function add(block)
  sname = get_param(block, 'signame');
  stype = get_param(block, 'sigtype');

  udata = get_param(block, 'userdata');
  switch stype
    case 'size'
      ssize = 1;
    otherwise
      ssize = udata.sigsize(sname);
  end
  udata.siglist = [udata.siglist ...
                     struct('name', sname, 'type', stype, 'size', ssize)];
  set_param(block, 'userdata', udata);

  update(block)
end

function remove(block)
  sname = get_param(block, 'signame');
  stype = get_param(block, 'sigtype');

  udata = get_param(block, 'userdata');
  udata.siglist = udata.siglist(~(strcmp({udata.siglist.name}, sname) & ...
                                  strcmp({udata.siglist.type}, stype)));
  set_param(block, 'userdata', udata);

  update(block)
end

function timetype(block)
  ttype = get_param(block, 'timetype');
  mask = Simulink.Mask.get(block);
  i = find(strcmp({mask.Parameters.Name}, 'period'));

  switch ttype
    case 'periodic'
      mask.Parameters(i).Enabled = 'on';
    otherwise
      mask.Parameters(i).Enabled = 'off';
  end
end

% --- signames -------------------------------------------------------------
%
% Retrieve signal names from meta description
%
function [names, sizes] = signames(format)
  names = {};
  sizes = containers.Map();
  for f = fieldnames(format)'
    [fnames, fsizes] = signames_1('', f{1}, format.(f{1}));
    names = [names fnames];
    sizes = [sizes ; fsizes];
  end
end

function [names, sizes] = signames_1(prefix, name, format)
  switch format.kind
    case {'struct' 'union'}
      key = [prefix name];
      names = { key };
      sizes = containers.Map(names, 0);
      for f = fieldnames(format.type)'
        [fnames, fsizes] = signames_1([key '.'], f{1}, format.type.(f{1}));
        names = [names fnames];
        sizes = [sizes ; fsizes];
        sizes(key) = sizes(key) + fsizes([key '.' f{1}]);
      end

    case {'array' 'sequence'}
      [names, sizes] = signames_1(prefix, name, format.type);
      if isfield(format, 'length')
        l = format.length;
      else
        l = 16;
      end
      for n = names
        sizes(n{1}) = l * sizes(n{1});
      end

    case 'optional'
      [names, sizes] = signames_1(prefix, name, format.type);

    otherwise
      names = { [prefix name] };
      sizes = containers.Map(names, 1);
  end
end
